/* DDS Function Generator
   John Richards 2021
   Speedy version -- using bit-shifts in the accumulator
   Based on code by Elliot Williams
   https://hackaday.com/2016/02/12/embed-with-elliot-audio-playback-with-direct-digital-synthesis/
   https://github.com/hexagon5un/barking_dogs_dds_demo
   Sine, triangle, saw and square waveforms (see dds_waveforms.h)
   Analogue read - frequency of oscillator (A0)
   Output (Pin 11)
   Pushbutton select waveform (Pin 2)
   Low-pass RC filter required, for example, 10k/0.1uF and output cap 47uF
*/

#include <avr/interrupt.h>
#include <avr/io.h>
#include <stdlib.h>
#include <avr/pgmspace.h>
#include "dds_waveforms.h"

#define clear_bit(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#define set_bit(sfr, bit)   (_SFR_BYTE(sfr) |= _BV(bit))

#define SAMPLE_RATE       24000UL //'UL' to force the constant into an unsigned long constant
#define NUM_WAVES        4

// select waveform
// 0 = Sine
// 1 = Triangle
// 2 = Saw
// 3 = Square
static uint8_t i = 0; //int to select waveforms (see setup)

// set pin numbers
int ledPin = 13;      // the number of the LED pin
int ledState = LOW;         // the current state of the output pin

int sensorPin = A0;    // select the input pin for the potentiometer
int sensorValue = 0;  // variable to store the value coming from the sensor

int buttonPin = 2;     // the number of the pushbutton pin
int buttonState;             // the current reading from the input pin
int lastButtonState = LOW;   // the previous reading from the input pin

// the following variables are unsigned longs because the time, measured in
// milliseconds, will quickly become a bigger number than can be stored in an int.
unsigned long lastDebounceTime = 0;  // the last time the output pin was toggled
unsigned long debounceDelay = 50;    // the debounce time; increase if the output flickers

int ctr = 0;

/* Init functions, defined at the bottom of the file */
static inline void setup_pwm_audio_timer(void);
static inline void setup_sample_timer(void);

struct DDS {
  uint16_t increment;
  uint16_t accumulator;
  const int8_t* sample;   /* pointer to beginning of sample in memory */
};

volatile struct DDS wave[NUM_WAVES];

ISR(TIMER1_COMPA_vect) {
  int16_t total = 0;

  total += (int8_t) pgm_read_byte_near(wave[i].sample + (wave[i].accumulator >> 8));
  wave[i].accumulator += wave[i].increment;

  OCR2A = total + 128; // add in offset to make it 0-255 rather than -128 to 127
}

void setup() {
  pinMode(buttonPin, INPUT);
  pinMode(ledPin, OUTPUT);

  // set initial LED state
  digitalWrite(ledPin, ledState);

  setup_sample_timer();
  setup_pwm_audio_timer();

  // waveforms (see dds_waveforms.h)
  wave[0].sample = Sine; // 0
  wave[1].sample = Triangle; // 1
  wave[2].sample = Saw; // 2
  wave[3].sample = Square; // 3
}

void loop() {
  unsigned long sensorValue = analogRead(A0); // unsigned long for squaring and mapping

  //  wave[i].increment = 2048; // fixed pitch

  //  some mapping see https://forum.arduino.cc/index.php?topic=145443.0
  sensorValue = sensorValue * sensorValue;  // 0-1046529 (squared). For log/curve
  wave[i].increment = sensorValue / 110;  // mapped 0-9513

  // wave[i].increment = map(sensorValue, 0, 1023, 0, 9600); // linear mapping

  // read the state of the switch into a local variable:
  int reading = digitalRead(buttonPin);

  // check to see if you just pressed the button
  // (i.e. the input went from LOW to HIGH), and you've waited long enough
  // since the last press to ignore any noise:

  // If the switch changed, due to noise or pressing:
  if (reading != lastButtonState) {
    // reset the debouncing timer
    lastDebounceTime = millis();
  }
  if ((millis() - lastDebounceTime) > debounceDelay) {
    // whatever the reading is at, it's been there for longer than the debounce
    // delay, so take it as the actual current state:

    // if the button state has changed:
    if (reading != buttonState) {
      buttonState = reading;

      if (buttonState == HIGH) {
        ctr++;
      }
      if (ctr >= 4) {
        ctr = 0;
      }
      //
      // wavetable select
      if (ctr == 0) {
        i = 0; // Sine
      }
      if (ctr == 1) {
        i = 1; // Triangle
      }
      if (ctr == 2) {
        i = 2; // Saw
      }
      if (ctr == 3) {
        i = 3; // Square
      }  
    }
  }
  // save the reading. Next time through the loop, it'll be the lastButtonState:
  lastButtonState = reading;
}

static inline void setup_sample_timer() {
  // Undo Arduino defaults:
  TCCR1A = 0;
  TCCR1B = 0;
  TIMSK1 = 0;
  // Set up in count-till-clear mode
  set_bit(TCCR1B, WGM12);
  // Sync to CPU clock: fast
  set_bit(TCCR1B, CS10);
  // Enable output-compare interrupt
  set_bit(TIMSK1, OCIE1A);
  OCR1A = F_CPU / SAMPLE_RATE - 1;
  sei();
}

static inline void setup_pwm_audio_timer() {
  // Undo Arduino defaults:
  TCCR2A = 0;
  TCCR2B = 0;
  // Set fast PWM mode
  TCCR2A |= _BV(WGM21) | _BV(WGM20);
  // With output on OC2A / Arduino pin 11
  set_bit(TCCR2A, COM2A1);
  // Set fastest clock speed: ~62kHz @ 16MHz
  set_bit(TCCR2B, CS20);
  // Signal is symmetric around 127
  OCR2A = 127;
  // output on pin 11 -- OC2A
  set_bit(DDRB, PB3);  /* or pinMode(11, OUTPUT) */
}
